from .middleware import MiddlewareManager
from .extension import ExtensionsManager, BaseExtension

from .handler import Handler
from .rule import RuleFactory
from .config import AbstractDispatcherConfig

import typing
import logging
import asyncio

logger = logging.getLogger(__name__)


class Dispatcher:
    def __init__(self, loop: asyncio.AbstractEventLoop, config: AbstractDispatcherConfig):
        self.loop: asyncio.AbstractEventLoop = loop
        self._hanlders: typing.List[Handler] = []
        self._config: AbstractDispatcherConfig = config

        self._middleware_manager: MiddlewareManager = MiddlewareManager(self)
        self._rule_factory: RuleFactory = RuleFactory({})
        self._extensions_manager: ExtensionsManager = ExtensionsManager(
            self, {}
        )

        self.__call_handler_logic: typing.Optional[typing.Callable] = None

    def register_call_handler_logic(self, call_handler_logic: typing.Callable):
        self.__call_handler_logic = call_handler_logic

    def _register_handler(self, handler: Handler):
        """
        Append handler to handlers list
        :param handler:
        :return:
        """
        self._hanlders.append(handler)
        logger.debug(f"Handler '{handler.handler.__name__}' successfully added!")

    def register_event_handler(
            self, coro: typing.Callable, event_type: AbstractDispatcherConfig.event_enum, rules: typing.List
    ):
        """
        Register event handler.

        >>> dp.register_event_hanlder(my_handler, Event.MyEvent, [])

        :param coro:
        :param event_type:
        :param rules:
        :return:
        """
        handler = Handler(event_type, coro, rules=rules)
        self._register_handler(handler)

    def event_handler(self, event_type: AbstractDispatcherConfig.event_enum, *rules, **named_rules):
        """
        Register event handler with decorator.

        @dp.event_handler(Event.MyEvent)
        async def my_func(event: eventobj.MyEvent, data: dict):
            print(event)

        :param event_type:
        :param rules:
        :param named_rules:
        :return:
        """

        def decorator(coro: typing.Callable):
            nonlocal named_rules
            named_rules = self._rule_factory.get_rules(named_rules)
            self.register_event_handler(coro, event_type, named_rules + list(rules))
            return coro

        return decorator

    def setup_middleware(self, middleware):
        """
        Add middleware to middlewares list with middleware manager.
        :param middleware:
        :return:
        """
        self._middleware_manager.setup(middleware)

    def setup_rule(self, rule):
        """
        Add named rule to named rules list with rule factory.
        :param rule:
        :return:
        """
        self._rule_factory.setup(rule)

    def setup_extension(self, extension: BaseExtension):
        self._extensions_manager.setup(extension)

    async def run_extension(self, extension_name: str, **kwargs):
        """

        :param extension_name: name of extension
        :param kwargs: named args which extension __init__ accepted.
        :return:
        """
        await self._extensions_manager.run_extension(extension_name, **kwargs)

    async def _process_event(self, event: dict):
        """
        Handle 1 event coming from extension.
        :param event:
        :return:
        """
        data = {}  # dict for transfer data from middlewares to handler (and rules).

        _skip_handler, data = await self._middleware_manager.trigger_pre_process_middlewares(
            event, data
        )  # trigger pre_process_event funcs in middlewares.
        # returns service value '_skip_handler' and data variable (check upper).

        if not _skip_handler:  # if middlewares don`t skip this handler, dispatcher be check rules and execute handlers.
            await self.__call_handler_logic(event, self._hanlders, self._config, data)

        await self._middleware_manager.trigger_post_process_middlewares()  # trigger post_process_event funcs in
        # middlewares.

    async def _process_events(self, events: typing.List[dict]):
        """
        Process events coming from extensions.
        :param events: list of events.
        :return:
        """
        for event in events:
            self.loop.create_task(self._process_event(event))

    async def _call_handler_logic(self, event: dict, handlers: typing.List[Handler], config: AbstractDispatcherConfig, data: dict):
        """
        Impelement your call handler logic here.
        :param config:
        :param data:
        :return:
        """
        raise NotImplementedError()
